import Foundation
import Combine

// used for simulating asynchronous work
public func sleepRandom() {
    Thread.sleep(forTimeInterval: .random(in: 0.2...1.0))
}
public let backgroundQueue = DispatchQueue(label: "Background", attributes: .concurrent)

// Example data types
public struct MyDataSource {
    public func getData(for userName: String) -> MyUserData {
        print("getting user data")
        sleepRandom()
        print("done getting user data")
        return .init()
    }
}
public struct MyUserData { }
public struct MyDataSourceFacade {
    let userData: MyUserData
    let A: MyDataSource
    let B: MyDataSource
    let C: MyDataSource

    public init(_ userData: MyUserData, _ A: MyDataSource, _ B: MyDataSource, _ C: MyDataSource) {
        self.userData = userData
        self.A = A
        self.B = B
        self.C = C
    }
}

// Example asynchronous tasks
public func getTokenFromServer() -> String {
    print("fetching token")
    sleepRandom()
    print("done fetching token")
    return "token"
}
public func getTokenPublisher() -> AnyPublisher<String, Error> {
    return Just("token")
        .delay(for: .init(Double.random(in: 0.2...1)), scheduler: RunLoop.main)
        .setFailureType(to: Error.self)
        .handleEvents { _ in
            print("fetching token")
        } receiveOutput: { _ in
            print("done fetching token")
        }
        .eraseToAnyPublisher()
}
public func getDataSourceA(_ token: String) -> MyDataSource {
    print("creating A")
    sleepRandom()
    print("done creating A")
    return .init()
}
public func getDataSourceB(_ userData: MyUserData, _ token: String) -> MyDataSource {
    print("creating B")
    sleepRandom()
    print("done creating B")
    return .init()
}
public func getDataSourceC(_ userData: MyUserData, _ token: String) -> MyDataSource {
    print("creating C")
    sleepRandom()
    print("done creating C")
    return .init()
}

// Conveniences for manipulating Combine publishers
public extension Publisher {
    /// Log upstream errors and don't pass failures down the pipeline
    func logError() -> AnyPublisher<Output?, Never> {
        return self
            .makeOptional()
            .mapError { error -> Failure in
                //print(error.localizedDescription)
                return error
            }
            .replaceError(with: nil)
            .eraseToAnyPublisher()
    }

    /// Wraps an upstream publisher's output in an Optional
    func makeOptional() -> AnyPublisher<Output?, Failure> {
        return self
            .map { input -> Output? in input }
            .eraseToAnyPublisher()
    }

    /// Filters out the nil values of an upstream with Optional Output, ensuring non-optional values downstream
    func ignoreNil<T>() -> AnyPublisher<T, Failure> where Output == Optional<T> {
        return self
            .compactMap { $0 }
            .eraseToAnyPublisher()
    }
}

